/*
 * ==========================
 * =       Proiect PM       =
 * ==========================
 * = TUDOR Cristian - 331CB =
 * ==========================
 */

#include "usart.h"

#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/io.h>


/* Constantele pentru comenzi */

#define FORWARD 'F'
#define BACK 'B'
#define LEFT 'L'
#define RIGHT 'R'
#define FORWARD_LEFT 'G'
#define FORWARD_RIGHT 'I'
#define BACK_LEFT 'H'
#define BACK_RIGHT 'J'
#define SPEED1 '1'
#define SPEED2 '2'
#define SPEED3 '3'
#define SPEED4 '4'
#define SPEED5 '5'
#define SPEED6 '6'
#define SPEED7 '7'
#define SPEED8 '8'
#define SPEED9 '9'
#define SPEED10 'q'

#define SENT 0
#define RECV 1
#define SLOWDOWN1 125
#define SLOWDOWN2 30

volatile unsigned long int distance = 0;
volatile unsigned long int count = 0;
volatile unsigned int message = 0;

int speed = -1;
int dist = 0;
int repeat = 0;

/*
 * Functie de activare a motorului pentru mers inainte
 */
void forward(void) {
    PORTB &= ~(1 << PB4);
	PORTB |= (1 << PB5);
}

/*
 * Functie de activare a motorului pentru mers inapoi
 */
void back(void) {
	PORTB |= (1 << PB4);
	PORTB &= ~(1 << PB5);
}

/*
 * Functie de activare a motorului pentru viraj stanga
 */
void left(void) {
	PORTB |= (1 << PB6);
	PORTB &= ~(1 << PB7);
}

/*
 * Functie de activare a motorului pentru viraj stanga
 */
void right(void) {
    PORTB &= ~(1 << PB6);
	PORTB |= (1 << PB7);
}

/*
 * Functie de dezactivare motoare
 */
void stop(void) {
	PORTB &= ~(1 << PB4);
	PORTB &= ~(1 << PB5);
	PORTB &= ~(1 << PB6);
	PORTB &= ~(1 << PB7);
}

/*
 * Functie de dezactivare a motorului pentru acceleratie
 */
void noAcceleration(void) {
    PORTB &= ~(1 << PB4);
	PORTB &= ~(1 << PB5);
}

/*
 * Functie de dezactivare a motorului pentru viraje
 */
void noDirection(void) {
    PORTB &= ~(1 << PB6);
	PORTB &= ~(1 << PB7);
}

/*
 * Functie de initializare motoare
 */
void motors_init(void) {
	DDRB |= (1 << PB4);
	DDRB |= (1 << PB5);
	DDRB |= (1 << PB6);
	DDRB |= (1 << PB7);

	stop();
}

/*
 * Initializare timer PWM pentru viteza variata a motorului din spate
 */
void init_timer0(void)
{
    DDRB |= (1 << PB3);
    TCCR0A |= (1 << WGM01) | (1 << WGM00) | (1 << COM0A1);
    OCR0A = 255;
    TCCR0B |= (1 << CS00) | (1 << CS02);
}

/*
 * Initializare timer 2 pentru senzorul cu ultrasunete HC-SR04
 */
void init_timer2(void)
{
	OCR2A = 16;
	TCCR2A = (1 << WGM21);
	TCCR2B = (1 << CS21) | (1 << CS20);
	TIMSK2 = (1 << OCIE2A);
}

/*
 * Initializare senzor cu ultrasunete HC-SR04
 */
void init_distance() {
    EIMSK |= (1 << INT0) | (1 << INT1);
	EICRA |= (1 << ISC00) | (1 << ISC10);

	DDRD |= (1 << PD4);
	DDRD &= ~(1 << PD3);
	PORTD |= (1 << PD3);

	DDRD |= (1 << PD5);
	DDRD &= ~(1 << PD2);
	PORTD |= (1 << PD2);
}

/*
 * Counter intrerupere
 */
ISR(TIMER2_COMPA_vect)
{
	count++;
}

/*
 * Intrerupere senzor SR04 din fata
 */
ISR(INT1_vect) {
	if((PIND & (1 << PD3)) == 0 ) {
		message = RECV;
		distance = count;
	} else {
        count = 0;
	}
}

/*
 * Intrerupere senzor SR04 din spate
 */
ISR(INT0_vect) {
	if((PIND & (1 << PD2)) == 0 ) {
		message = RECV;
		distance = count;
	} else {
        count = 0;
	}
}

/*
 * Determinare distanta
 */
int get_distance(char type)
{
	message = SENT;
	distance = 0;

	if (type == FORWARD) {
        PORTD &= ~(1 << PD4);
        _delay_us(2);

        PORTD |= (1 << PD4);
        _delay_us(10);
        PORTD &= ~(1 << PD4);
	} else {
        PORTD &= ~(1 << PD5);
        _delay_us(2);

        PORTD |= (1 << PD5);
        _delay_us(10);
        PORTD &= ~(1 << PD5);
	}

	while(message != RECV);

	return distance;
}

/*
 * Functie de deplasare in fata daca senzorul nu detecteaza nimic in cale
 */
int go_forward(char received_char) {
    dist = get_distance(FORWARD);

    if (dist < SLOWDOWN1 && (received_char == FORWARD ||
        received_char == FORWARD_LEFT || received_char == FORWARD_RIGHT)) {
        if (dist < SLOWDOWN2) {
            stop();
            return 0;
        }

        if (speed == -1)
            speed = OCR0A;
        OCR0A = 125;
        repeat = 1;
    }

    forward();
    return 1;
}

/*
 * Functie de deplasare in fata-stanga daca senzorul nu detecteaza nimic in cale
 */
void go_forward_left(char received_char) {
    if (go_forward(received_char) != 0)
        left();
}

/*
 * Functie de deplasare in fata-dreapta daca senzorul nu detecteaza nimic in cale
 */
void go_forward_right(char received_char) {
    if (go_forward(received_char) != 0)
        right();
}

/*
 * Functie de deplasare in spate daca senzorul nu detecteaza nimic in cale
 */
int go_back(char received_char) {
    dist = get_distance(BACK);

    if (dist < SLOWDOWN1 && (received_char == BACK ||
        received_char == BACK_LEFT || received_char == BACK_RIGHT)) {
        if (dist < SLOWDOWN2) {
            stop();
            return 0;
        }

        if (speed == -1)
            speed = OCR0A;
        OCR0A = 125;
        repeat = 1;
    }

    back();
    return 1;
}

/*
 * Functie de deplasare in spate-stanga daca senzorul nu detecteaza nimic in cale
 */
void go_back_left(char received_char) {
    if (go_back(received_char) != 0)
        left();
}

/*
 * Functie de deplasare in spate-dreapta daca senzorul nu detecteaza nimic in cale
 */
void go_back_right(char received_char) {
    if (go_back(received_char) != 0)
        right();
}

int main(void) {

	char received_char;

	USART0_init();
	motors_init();
	init_distance();
	init_timer0();
	init_timer2();
	sei();

	while(1) {
        repeat = 0;
		received_char = USART0_receive();

		switch (received_char) {
        case FORWARD:
            noDirection();
            go_forward(received_char);
            break;
        case BACK:
            noDirection();
            go_back(received_char);
            break;
        case LEFT:
            noAcceleration();
            left();
            break;
        case RIGHT:
            noAcceleration();
            right();
            break;
        case FORWARD_LEFT:
            go_forward_left(received_char);
            break;
        case FORWARD_RIGHT:
            go_forward_right(received_char);
            break;
        case BACK_LEFT:
            go_back_left(received_char);
            break;
        case BACK_RIGHT:
            go_back_right(received_char);
            break;

        case SPEED1:
            OCR0A = 125;
            break;
        case SPEED2:
            OCR0A = 140;
            break;
        case SPEED3:
            OCR0A = 155;
            break;
        case SPEED4:
            OCR0A = 170;
            break;
        case SPEED5:
            OCR0A = 185;
            break;
        case SPEED6:
            OCR0A = 200;
            break;
        case SPEED7:
            OCR0A = 215;
            break;
        case SPEED8:
            OCR0A = 230;
            break;
        case SPEED9:
            OCR0A = 245;
            break;
        case SPEED10:
            OCR0A = 255;
            break;

        default:
            stop();
		}

		if (speed != -1 && repeat == 0) {
            OCR0A = speed;
            speed = -1;
        }

		_delay_ms(25);
	}

	return 0;
}
